MODULE wzvmod
   !!==============================================================================
   !!                       ***  MODULE  wzvmod  ***
   !! Ocean diagnostic variable : vertical velocity
   !!==============================================================================
   !! History :   5.0  !  90-10  (C. Levy, G. Madec)  Original code
   !!             7.0  !  96-01  (G. Madec)  Statement function for e3
   !!             8.5  !  02-07  (G. Madec)  Free form, F90
   !!              "   !  07-07  (D. Storkey) Zero zhdiv at open boundary (BDY) 
   !!----------------------------------------------------------------------
   !!   wzv        : Compute the vertical velocity
   !!----------------------------------------------------------------------
   !! * Modules used
   USE oce             ! ocean dynamics and tracers variables
   USE dom_oce         ! ocean space and time domain variables 
   USE sbc_oce         ! surface boundary condition: ocean
   USE domvvl          ! Variable volume
   USE in_out_manager  ! I/O manager
   USE prtctl          ! Print control
   USE phycst
   USE bdy_oce         ! unstructured open boundaries
   USE lbclnk          ! ocean lateral boundary condition (or mpp link)
   USE obc_par         ! open boundary cond. parameter
   USE obc_oce

   IMPLICIT NONE
   PRIVATE

   !! * Routine accessibility
   PUBLIC wzv          ! routine called by step.F90 and inidtr.F90
   PUBLIC ssh_wzv
   PUBLIC ssh_nxt

   !! * Substitutions
#  include "domzgr_substitute.h90"
   !!----------------------------------------------------------------------
   !!  OPA 9.0 , LOCEAN-IPSL (2005) 
   !! $Id$
   !! Software governed by the CeCILL licence (modipsl/doc/NEMO_CeCILL.txt)
   !!----------------------------------------------------------------------

CONTAINS

   SUBROUTINE wzv( kt )
      !!----------------------------------------------------------------------
      !!                    ***  ROUTINE wzv  ***
      !!
      !! ** Purpose :   Compute the now vertical velocity after the array swap
      !!
      !! ** Method  :   Using the incompressibility hypothesis, the vertical
      !!      velocity is computed by integrating the horizontal divergence 
      !!      from the bottom to the surface.
      !!        The boundary conditions are w=0 at the bottom (no flux) and,
      !!      in regid-lid case, w=0 at the sea surface.
      !!
      !! ** action  :   wn array : the now vertical velocity
      !!----------------------------------------------------------------------
      INTEGER, INTENT(in) :: kt
      
      ! Empty routine

      WRITE(*,*) 'wzv : you should not be here : error ?'  

   END SUBROUTINE wzv

   SUBROUTINE ssh_wzv( kt ) 
      !!----------------------------------------------------------------------
      !!                ***  ROUTINE dom_vvl_ssh  ***
      !!                   
      !! ** Purpose :   compute the after ssh (ssha), the now vertical velocity
      !!              Cand update the now vertical coordinate (lk_vvl=T).
      !!
      !! ** Method  : -  

      !!              - Using the incompressibility hypothesis, the vertical
      !!      velocity is computed by integrating the horizontal divergence 
      !!      from the bottom to the surface minus the scale factor evolution.
      !!        The boundary conditions are w=0 at the bottom (no flux) and.
      !!
      !! ** action  :   wn array : the now vertical velocity
      !!----------------------------------------------------------------------
      INTEGER, INTENT(in) ::   kt   ! time step
      !!
      INTEGER  ::   ji, jj, jk      ! dummy loop indices
      REAL(wp) ::   z2dt, zraur     ! temporary scalars
      REAL(wp), DIMENSION(jpi,jpj) ::   zhdiv       ! 2D workspace
      !!----------------------------------------------------------------------

      IF( kt == nit000 ) THEN
         IF(lwp) WRITE(numout,*)
         IF(lwp) WRITE(numout,*) 'ssh_wzv : vertical velocity from continuity eq. (vvl option)'
         IF(lwp) WRITE(numout,*) '~~~~~~~ '
         !
         wn(:,:,jpk) = 0.e0      ! bottom boundary condition: w=0 (set once for all)
      ENDIF

      ! set time step size (Euler/Leapfrog)
      z2dt = 2. * rdt 
      IF( neuler == 0 .AND. kt == nit000 ) z2dt =rdt

      zraur = 1. / rauw

      !                                           !------------------------------!
      !                                           !   After Sea Surface Height   !
      !                                           !------------------------------!
      zhdiv(:,:) = 0.e0
      DO jk = 1, jpkm1                                 ! Horizontal divergence of barotropic transports
        zhdiv(:,:) = zhdiv(:,:) + fse3t(:,:,jk) * hdivn(:,:,jk)
      END DO

      !                                                ! Sea surface elevation time stepping
      ssha(:,:) = (  sshb(:,:) - z2dt * ( zraur * emp(:,:)                       + zhdiv(:,:) )  ) * tmask(:,:,1)

      !                                                ! Sea Surface Height at u-,v- and f-points (vvl case only)
      IF( lk_vvl ) THEN                                ! (required only in key_vvl case)
!RB_vvl
! Compute ssh at u,v, f points and update vertical coordinate
! Currently done in dom_vvl
      ENDIF

      !                                           !------------------------------!
      !                                           !     Now Vertical Velocity    !
      !                                           !------------------------------!
      !                                                ! integrate from the bottom the hor. divergence
         DO jk = jpkm1, 1, -1
            wn(:,:,jk) = wn(:,:,jk+1) -    fse3t_n(:,:,jk) * hdivn(:,:,jk)   &
                 &                    - (  fse3t_a(:,:,jk)                   &
                 &                       - fse3t_b(:,:,jk) ) * tmask(:,:,jk) / z2dt
         END DO

      !                                           !------------------------------!
      !                                           !  Update Now Vertical coord.  !
      !                                           !------------------------------!
      IF( lk_vvl ) THEN                           ! only in vvl case)
      !                                                ! now local depth and scale factors (stored in fse3. arrays)
!RB_vvl to be done


      ENDIF
      !
   END SUBROUTINE ssh_wzv


   SUBROUTINE ssh_nxt( kt )
      !!----------------------------------------------------------------------
      !!                    ***  ROUTINE ssh_nxt  ***
      !!
      !! ** Purpose :   achieve the sea surface  height time stepping by 
      !!              applying Asselin time filter and swapping the arrays
      !!              ssha  already computed in ssh_wzv  
      !!
      !! ** Method  : - apply Asselin time fiter to now ssh and swap :
      !!             sshn = ssha + atfp * ( sshb -2 sshn + ssha )
      !!             sshn = ssha
      !!
      !! ** action  : - sshb, sshn   : before & now sea surface height
      !!                               ready for the next time step
      !!----------------------------------------------------------------------
      INTEGER, INTENT( in ) ::   kt      ! ocean time-step index
      !!
      INTEGER  ::   ji, jj, jk           ! dummy loop indices
      !!----------------------------------------------------------------------

      IF( kt == nit000 ) THEN
         IF(lwp) WRITE(numout,*)
         IF(lwp) WRITE(numout,*) 'ssh_nxt : next sea surface height (Asselin time filter + swap)'
         IF(lwp) WRITE(numout,*) '~~~~~~~ '
      ENDIF

      ! Time filter and swap of the ssh
      ! -------------------------------

!RB_vvl ssh at u, f,v point to be added

     IF( neuler == 0 .AND. kt == nit000 ) THEN   ! Euler time stepping
         DO jj = 1, jpj
            DO ji = 1, jpi
               ! before <-- now
               sshb  (ji,jj) = sshn(ji,jj) 
               ! now <-- after
               sshn  (ji,jj) = ssha  (ji,jj)
            END DO
         END DO
     ELSE                                         ! Leap-frog time stepping
         DO jj = 1, jpj
            DO ji = 1, jpi
               ! before <-- now filtered
               sshb  (ji,jj) = sshn(ji,jj)   + atfp * ( sshb  (ji,jj) - 2 * sshn  (ji,jj) + ssha  (ji,jj) )    !&
               ! now <-- after
               sshn  (ji,jj) = ssha  (ji,jj)
            END DO
         END DO
      ENDIF

      IF(ln_ctl)   CALL prt_ctl(tab2d_1=sshb    , clinfo1=' sshb  - : ', mask1=tmask, ovlap=1 )
      !
   END SUBROUTINE ssh_nxt

   !!======================================================================
END MODULE wzvmod
